angular.module('PasswordMonitor').factory('IdbService', function ($q, HashingService) {

    var idb = null;

    function getDB() {
        var deferred = $q.defer();

        if (idb !== null) {
            deferred.resolve(idb);
        }
        else if ("indexedDB" in window) {

            var openRequest = indexedDB.open('passwordMonitor', 75);

            openRequest.onupgradeneeded = function () {

                var thisDB = openRequest.result;
                if (!thisDB.objectStoreNames.contains('metrics')) {
                    thisDB.createObjectStore('metrics', {keyPath: 'passwordHash'});
                }
                if (!thisDB.objectStoreNames.contains('passwordVariations')) {
                    thisDB.createObjectStore('passwordVariations', {
                        unique: true, keyPath: [
                            'passwordHash',
                            'variationHash'
                        ]
                    });
                }
                if (!thisDB.objectStoreNames.contains('credentialUses')) {
                    var credentialUsesStore = thisDB.createObjectStore('credentialUses', {
                        keyPath: [
                            'usernameHash',
                            'passwordHash',
                            'url',
                            'time'
                        ]
                    });
                    credentialUsesStore.createIndex('passwordHash', 'passwordHash', {unique: false});
                    credentialUsesStore.createIndex('time', 'time', {unique: true});
                    credentialUsesStore.createIndex('url', 'url', {unique: false});
                }
                if (!thisDB.objectStoreNames.contains('deletedEntries')) {
                    thisDB.createObjectStore('deletedEntries', {keyPath: 'timeDeleted'});
                }
                if (!thisDB.objectStoreNames.contains('oauth')) {
                    thisDB.createObjectStore('oauth', {keyPath: 'time'});
                }
            };

            openRequest.onblocked = function () {
                console.error("IndexedDB blocked");
                deferred.reject();
            };

            openRequest.onerror = function (error) {
                console.error("Open db request error: ", error);
                deferred.reject();
            };

            openRequest.onsuccess = function () {
                idb = openRequest.result;
                deferred.resolve(idb);
            }
        }
        else {
            console.error("No IndexedDB support!");
            deferred.reject();
        }

        return deferred.promise;
    }

    /**
     * When we have deleted entries from the credentialUses table, it's important that we check that
     * we don't have remaining corresponding entries in the metrics table. We shouldn't keep metrics about
     * password uses that have been deleted, unless the password has also been used at other sites.
     */
    function checkConsistency(metricsStore, credentialUsesStore) {

        var cursorRequest = metricsStore.openCursor();
        var passwordIndex = credentialUsesStore.index('passwordHash');

        cursorRequest.onsuccess = function () {

            var cursor = cursorRequest.result;
            if (cursor) {
                var passwordHash = cursor.value['passwordHash'];
                var indexReq = passwordIndex.get(passwordHash);
                indexReq.onerror = function (error) {
                    console.error("Could not get value in index: ", error);
                    cursor.continue();
                };
                indexReq.onsuccess = function () {
                    if (indexReq.result === undefined) {
                        var deletePasswordMetricsEntryRequest = cursor.delete();
                        deletePasswordMetricsEntryRequest.onerror = function (error) {
                            console.error("Could not delete uncorresponding password metrics entry: ", error);
                            console.error(cursor.value);
                        }
                    }
                    cursor.continue();
                };

            }
        };

        cursorRequest.onerror = function (error) {
            console.error("Could not get entries from DB: ", error);
        };
    }

    function deleteEntry(entry, metricsStore, credUsesStore, deletedStore) {
        var getEntryToDeleteRequest = credUsesStore.get([
            entry['usernameHash'],
            entry['passwordHash'],
            entry['url'],
            entry['time']
        ]);
        getEntryToDeleteRequest.onsuccess = function () {
            var url = getEntryToDeleteRequest.result['url'];
            HashingService.hashThis(url).then(function (hashedUrl) {
                var addDeletedEntryRequest = deletedStore.add({
                    timeDeleted: new Date().getTime(),
                    time: getEntryToDeleteRequest.result['time'],
                    hashedUrl: hashedUrl
                });
                addDeletedEntryRequest.onsuccess = function () {
                    var deleteRequest = credUsesStore.delete([
                        entry['usernameHash'],
                        entry['passwordHash'],
                        entry['url'],
                        entry['time']
                    ]);
                    deleteRequest.onsuccess = function () {
                        checkConsistency(metricsStore, credUsesStore);
                    };
                    deleteRequest.onerror = function (error) {
                        console.error("Could not delete entry with username: %s, password: %s and URL: %s from DB",
                            entry['usernameHash'], entry['passwordHash'], entry['url']);
                        console.error(error);
                    };
                };
                addDeletedEntryRequest.onerror = function (error) {
                    console.error("Could not transfer entry into deletedEntries DB: ", error);
                };
            });
        };
        getEntryToDeleteRequest.onerror = function (error) {
            console.error("Could not find entry to delete in DB: ", error);
        }
    }

    function getAllValues(table) {
        var deferred = $q.defer();
        getDB().then(function (idb) {
            var transaction = idb.transaction([table], 'readonly');
            var store = transaction.objectStore(table);
            var cursor = store.openCursor();

            var entriesToReturn = [];
            cursor.onsuccess = function () {
                var result = cursor.result;
                if (result) {
                    entriesToReturn.push(result.value);
                    result.continue();
                }
                else {
                    deferred.resolve(entriesToReturn);
                }
            };

            cursor.onerror = function (error) {
                console.error("Could not get entries from DB");
                deferred.reject(error);
            };
        }, function (error) {
            deferred.reject(error);
        });

        return deferred.promise;
    }

    function getLastEntryWithUrl(url, time) {
        var deferred = $q.defer();
        getDB().then(function (idb) {
            var transaction = idb.transaction(['credentialUses'], 'readonly');
            var credentialUsesStore = transaction.objectStore('credentialUses');
            var timeIndex = credentialUsesStore.index('time');
            var cursorRequest = timeIndex.openCursor(IDBKeyRange.lowerBound(time - 600000), "prev"); // 600000 = 10 min
            var prevTry = null;
            cursorRequest.onsuccess = function () {
                var cursor = cursorRequest.result;
                if (cursor) {
                    if (!prevTry && cursor.value.url === url) {
                        prevTry = cursor.primaryKey;
                    }
                    cursor.continue();
                }
                else {
                    deferred.resolve(prevTry);
                }
            };
            cursorRequest.onerror = function (error) {
                console.error(error);
                deferred.reject(error);
            }
        });
        return deferred.promise;
    }

    function hasPasswordAlready(passwordHash) {
        var deferred = $q.defer();
        getDB().then(function (idb) {
            var transaction = idb.transaction(['metrics'], 'readonly');
            var metricsStore = transaction.objectStore('metrics');
            var metricsGetRequest = metricsStore.get(passwordHash);
            metricsGetRequest.onsuccess = function () {
                var result = metricsGetRequest.result;
                deferred.resolve(result !== undefined);
            };
            metricsGetRequest.onerror = function (error) {
                console.error(error);
                deferred.reject(error);
            }
        });
        return deferred.promise;
    }

    return {

        getAllValues: getAllValues,

        hasPasswordAlready: hasPasswordAlready,

        insertNewPasswordEntry: function (usernameHash, passwordHash, url, time, timeSpent, length, numOfCharClasses,
                                          structure, entropy, crackTime, variations) {

            if (usernameHash && passwordHash && url && time && timeSpent && length && numOfCharClasses &&
                structure && entropy && crackTime && variations) {

                getDB().then(function (idb) {
                    var transaction = idb.transaction(['metrics'], 'readwrite');

                    // Add all the metrics info to the DB
                    var metricsStore = transaction.objectStore('metrics');
                    var newMetricsEntry = {
                        passwordHash: passwordHash,
                        length: length,
                        numOfCharClasses: numOfCharClasses,
                        structure: structure,
                        entropy: entropy,
                        crackTime: crackTime
                    };
                    metricsStore.add(newMetricsEntry).onerror = function (error) {
                        console.error("Could not add new entry to metrics table: ", error);
                        console.error(newMetricsEntry);
                    };

                    // Add all the different password variations to the DB
                    transaction = idb.transaction(['passwordVariations'], 'readwrite');
                    var variationsStore = transaction.objectStore('passwordVariations');
                    for (var n = 0; n < variations.length; n++) {
                        (function (n) {
                            var newPasswordVariationsEntry = {
                                passwordHash: passwordHash,
                                variationHash: variations[n]['variationHash'],
                                reason: variations[n]['reason']
                            };
                            var addRequest = variationsStore.add(newPasswordVariationsEntry);
                            addRequest.onerror = function (error) {
                                console.error("Could not add entry to passwordVariations table: %s",
                                    addRequest.error.message);
                                console.error(error);
                            };
                        })(n);
                    }
                });
            }
        },

        insertNewUsageEntry: function (usernameHash, passwordHash, url, time, timeSpent) {
            if (usernameHash && passwordHash && url && time && timeSpent) {
                $q.all([getDB(), getLastEntryWithUrl(url, time)]).then(function (results) {
                    var transaction = results[0].transaction(['credentialUses'], 'readwrite');
                    var credentialUsesStore = transaction.objectStore('credentialUses');
                    var newCredentialUsesEntry = {
                        usernameHash: usernameHash,
                        passwordHash: passwordHash,
                        url: url,
                        time: time,
                        timeSpent: timeSpent,
                        prevRetry: results[1]
                    };
                    var credentialUsesAddRequest = credentialUsesStore.add(newCredentialUsesEntry);
                    credentialUsesAddRequest.onerror = function (error) {
                        console.error("Could not add entry to credential uses table: ", error);
                        console.error(newCredentialUsesEntry);
                    };
                });
            }
        },

        insertNewOAuthEntry: function (oauthProviderUrl, url, time) {
            if (oauthProviderUrl && url && time) {
                getDB().then(function (idb) {
                    var transaction = idb.transaction(['oauth'], 'readwrite');
                    var oauthStore = transaction.objectStore('oauth');
                    var newOAuthEntry = {
                        oauthProviderUrl: oauthProviderUrl,
                        url: url,
                        time: time
                    };
                    var oauthAddRequest = oauthStore.add(newOAuthEntry);
                    oauthAddRequest.onerror = function (error) {
                        console.error("Could not add entry to OAuth table:", error);
                        console.error(newOAuthEntry);
                    };
                });
            }
        },

        deleteCheckedEntries: function (currentEntries, callback) {
            if (currentEntries && currentEntries.length) {
                getDB().then(function (idb) {
                    var transaction = idb.transaction(['metrics', 'credentialUses', 'deletedEntries'], 'readwrite');
                    var metricsStore = transaction.objectStore('metrics');
                    var credentialUsesStore = transaction.objectStore('credentialUses');
                    var deletedEntriesStore = transaction.objectStore('deletedEntries');
                    angular.forEach(currentEntries, function (entry) {
                        if (entry.checked) {
                            deleteEntry(entry, metricsStore, credentialUsesStore, deletedEntriesStore);
                        }
                    });
                    transaction.oncomplete = function () {
                        callback();
                    }
                });
            }
            else {
                callback();
            }
        },

        deleteEntriesFromUrl: function (url) {
            if (url) {
                getDB().then(function (idb) {
                    var transaction = idb.transaction(['metrics', 'credentialUses', 'deletedEntries'], 'readwrite');
                    var metricsStore = transaction.objectStore('metrics');
                    var credentialUsesStore = transaction.objectStore('credentialUses');
                    var deletedEntriesStore = transaction.objectStore('deletedEntries');
                    var cursor = credentialUsesStore.openCursor();
                    cursor.onsuccess = function () {
                        var result = cursor.result;
                        if (result) {
                            if (result.value['url'].match(url)) {
                                deleteEntry(result.value, metricsStore, credentialUsesStore, deletedEntriesStore);
                            }
                            result.continue();
                        }
                    };
                });
            }
        },

        getAllTables: function () {
            var deferred = $q.defer();
            getDB().then(function (idb) {
                var qs = [];
                angular.forEach(idb['objectStoreNames'], function (store) {
                    qs.push(getAllValues(store));
                });
                return $q.all(qs);
            }).then(function (results) {
                    var tables = {};
                    for (var x = 0; x < idb['objectStoreNames'].length; x++) {
                        tables[idb['objectStoreNames'][x]] = results[x];
                    }
                    deferred.resolve(tables);
                },
                function () {
                    deferred.reject();
                }
            );
            return deferred.promise;
        },

        hasCollectedFromThisSite: function (url) {
            var deferred = $q.defer();
            getDB().then(function (idb) {
                var transaction = idb.transaction(['credentialUses'], 'readonly');
                var credentialUsesStore = transaction.objectStore('credentialUses');
                var urlIndex = credentialUsesStore.index('url');
                var urlGetRequest = urlIndex.get(url);
                urlGetRequest.onsuccess = function () {
                    var result = urlGetRequest.result;
                    deferred.resolve(angular.isDefined(result));
                };
            });
            return deferred.promise;
        }
    };
});
